#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 18 15:52:01 2020
modeling operation.
requirements = [
    'matplotlib>=3.1.3',
    'sklearn>=0.22.1',
    'seaborn>=0.10.0',
    'factor_analyzer>=0.3.2',
    'joblib>=0.14.1',
]
@author: zoharslong
"""
from numpy import max as np_max, min as np_min, ravel as np_ravel, argsort as np_argsort, abs as np_abs
from numpy import array as np_array, int32 as np_int32, int64 as np_int64, float64 as np_float64
from pandas import DataFrame, concat as pd_concat
from re import findall as re_find
from random import sample as rnd_smp
from os.path import join as os_join
from pyzohar.sub_slt_bsc.bsz import lsz
from pyzohar.sub_slt_bsc.dfz import dfz
import seaborn as sns  # plot on factor analysis
from factor_analyzer import FactorAnalyzer, calculate_kmo, calculate_bartlett_sphericity, Rotator
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.cluster import KMeans
from sklearn.svm import SVC, SVR
from sklearn.svm._classes import SVC as typ_svc, SVR as typ_svr
from sklearn.manifold import TSNE
from matplotlib import pyplot as plt
from joblib import dump as jb_dump, load as jb_load


class mdz_nit(dfz):
    """
    modeling class initiation
    >>> mdl = mdz_nit(DataFrame(),dfz_mpt=dfz(),dct_mdl={'_id':'','fld':'','clm_x':[],'clm_y':[],'drp':None,'kep':None})
    >>> mdl.mdl_nit()
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        :param dfz_mpt: a dfz for inserted
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param dct_mdl: more params are auto generated, in ['fls', 'smp_raw/_y_0/_y_1/_trn/_tst']
        """
        if dfz_mpt is not None:
            dts = dts if dts is not None else dfz_mpt.dts.copy()
            lcn = lcn if lcn is not None else dfz_mpt.lcn.copy()
        super(mdz_nit, self).__init__(dts, lcn, spr=spr)
        self.clm_x, self.clm_y, self._x, self._y, self.mdl = None, None, None, None, None
        self.mdl_nit(dct_mdl=dct_mdl)

    def mdl_nit(self, dct_mdl=None, rst=True):
        """
        针对model的初始化, 会使用self.dts对self.lcn.mdl的一系列params进行刷新, 通常在self.dts发生变化的时候必须调用
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param rst: reset self.lcn.mdl or not, default True
        :return: None
        """
        if dct_mdl is not None:
            mch_tmp = [self.lcn['mdl']['_id'], self.lcn['mdl']['clm_y'], self.lcn['mdl']['clm_x']] if \
                'mdl' in self.lcn.keys() else None
            dct_mdl['clm_x'] = lsz(dct_mdl['clm_x']).typ_to_lst(rtn=True)
            dct_mdl['clm_y'] = lsz(dct_mdl['clm_y']).typ_to_lst(rtn=True)
            self.lcn.update({'mdl': dct_mdl})
            self.clm_x = self.lcn['mdl']['clm_x']
            self.clm_y = self.lcn['mdl']['clm_y']
            self.lcn['mdl']['fls'] = 'mdl_' + self.lcn['mdl']['_id'] + '_' + self.lcn['mdl']['clm_y'][0] + '.pkl'
            if mch_tmp != [self.lcn['mdl']['_id'], self.lcn['mdl']['clm_y'], self.lcn['mdl']['clm_x']]:
                self.mdl, self._x, self._y = None, None, None  # 当参数发生较大改变时, 删除已经存在的self.mdl
        if rst:
            self.mdl_smp_trc()

    def mdl_smp_trc(self):
        """
        basic sample trace
        """
        tmp = dfz(self.dts.copy(), self.lcn.copy())
        if tmp.len > 0:
            # 标记数据集筛选
            if self.lcn['mdl']['drp']:
                tmp.drp_dcm_ctt(self.lcn['mdl']['drp'], prm='drop', ndx_rst=False)
            if self.lcn['mdl']['kep']:
                tmp.drp_dcm_ctt(self.lcn['mdl']['kep'], prm='keep', ndx_rst=False)
            # 生成样本参数
            self.lcn['mdl']['smp_raw'] = {'ndx': tmp.dts.index.tolist(), 'clm_x': self.clm_x, 'clm_y': self.clm_y}
            tmp.drp_dcm_ctt({self.clm_y[0]: 1}, prm='keep', ndx_rst=False)
            self.lcn['mdl']['smp_y_1'] = {'ndx': tmp.dts.index.tolist(), 'clm_x': self.clm_x, 'clm_y': self.clm_y}
            dff_ndx = lsz().mrg('differ', self.lcn['mdl']['smp_raw']['ndx'], self.lcn['mdl']['smp_y_1']['ndx'],
                                rtn=True)
            self.lcn['mdl']['smp_y_0'] = {'ndx': dff_ndx, 'clm_x': self.clm_x, 'clm_y': self.clm_y}
            self._dl_smp_blc()  # generate smp_trn, smp_tst

    def _dl_smp_blc(self):
        """
        train/test sample balanced; 样本平衡策略, 以标志变量中较少的一方为基准, 保持标志平衡, 测试集占比10%
        >>> # 另一种既有的训练集划分方法
        >>> x_trn, x_tst, y_trn, y_tst = train_test_split(self._x, self._y, test_size=0.2, random_state=2)
        """
        smp_min = min(len(self.lcn['mdl']['smp_y_1']['ndx']), len(self.lcn['mdl']['smp_y_0']['ndx']))
        ndx_y_0 = rnd_smp(self.lcn['mdl']['smp_y_0']['ndx'], int(smp_min * 0.9))
        ndx_y_1 = rnd_smp(self.lcn['mdl']['smp_y_1']['ndx'], int(smp_min * 0.9))
        self.lcn['mdl']['smp_trn'] = {'ndx': ndx_y_0 + ndx_y_1, 'clm_x': self.clm_x, 'clm_y': self.clm_y}
        ndx_t_0 = lsz().mrg('differ', self.lcn['mdl']['smp_y_0']['ndx'], ndx_y_0, rtn=True)
        ndx_t_1 = lsz().mrg('differ', self.lcn['mdl']['smp_y_1']['ndx'], ndx_y_1, rtn=True)
        if len(self.lcn['mdl']['smp_y_1']['ndx']) < len(self.lcn['mdl']['smp_y_0']['ndx']):
            ndx_t_0 = rnd_smp(ndx_t_0, int(smp_min * 0.1))
        else:
            ndx_t_1 = rnd_smp(ndx_t_1, int(smp_min * 0.1))
        self.lcn['mdl']['smp_tst'] = {'ndx': ndx_t_0 + ndx_t_1, 'clm_x': self.clm_x, 'clm_y': self.clm_y}

    def mdl_smp_mpt(self, tgt=None, *, rtn=False):
        """
        model sample from param to real datasets, saved in self._x and self._y
        :param tgt: in [None, 'smp_raw', 'smp_trn', 'smp_tst']
        :param rtn: if return (dtf_y, dtf_x) or not, default False; if return, self._x/_y will not change
        :return: if rtn is True, return (dtf_y, dtf_x)
        """
        tmp, y, x = dfz(self.dts, self.lcn), None, None
        if tmp.len > 0:
            if tgt is None:     # 无样本集条件，则默认导入self.dts
                y, x = tmp.dts[self.clm_y], tmp.dts[self.clm_x]
            else:               # 否则根据样本集参数选择样本集
                tmp.drp_dcm(self.lcn['mdl'][tgt]['ndx'], prm='keep', ndx_rst=False)
                y, x = tmp.dts[self.lcn['mdl'][tgt]['clm_y']], tmp.dts[self.lcn['mdl'][tgt]['clm_x']]
        if rtn:
            return y, x
        self._y, self._x = y, x

    def mdl_sav(self, *, fld=None, fls=None):
        """
        save model into .pkl file
        :param fld: default None, get fold location from self.lcn
        :param fls: default None, get file location from self.lcn
        :return: None
        """
        fld = self.lcn['mdl']['fld'] if fld is None else fld
        fls = self.lcn['mdl']['fls'] if fls is None else fls
        jb_dump(self.mdl, os_join(fld, fls))

    def mdl_lod(self, *, fld=None, fls=None):
        """
        import model from .pkl file
        :param fld: default None, get fold location from self.lcn
        :param fls: default None, get file location from self.lcn
        :return: None
        """
        fld = self.lcn['mdl']['fld'] if fld is None else fld
        fls = self.lcn['mdl']['fls'] if fls is None else fls
        self.mdl = jb_load(os_join(fld, fls))
        str_tmp = self.lcn['mdl']['_id']
        self.lcn['mdl']['_id'] = re_find('mdl_(.*?)_' + self.lcn['mdl']['clm_y'][0], self.lcn['mdl']['fls'])[0]
        if str_tmp != self.lcn['mdl']['_id']:
            print('info: _id switch from %s to %s' % (str_tmp, self.lcn['mdl']['_id']))

    def drp_dcm_for_bly(self, flt_smp=2):
        """
        drop documents for the balance of Target label. 使用self.clm_y进行样本数量平衡, 会直接导致self.dts的改变
        :param flt_smp: 选择数量较多的label进行抽样，使其数量达到flt_smp倍于数量较少的label
        :return: None
        """
        # 不同分类的样本数量控制
        stm_1 = self.dts.loc[self.dts[self.clm_y[0]] == 1]
        stm_0 = self.dts.loc[self.dts[self.clm_y[0]] == 0]
        shp_0, shp_1 = stm_0.shape[0], stm_1.shape[0]
        try:
            stm_0 = stm_0.sample(int(shp_1 * flt_smp)) if shp_1*flt_smp < shp_0 else stm_0
            stm_1 = stm_1.sample(int(shp_0 * flt_smp)) if shp_0*flt_smp < shp_1 else stm_1
        except ValueError:
            pass
        self.dts = pd_concat([stm_1, stm_0])
        self.mdl_nit()


class mdz_fct(mdz_nit):
    """
    model factor analysis
    聚类的验证 https://blog.csdn.net/u010159842/article/details/78624135
    >>> mdl = mdz(DataFrame(),dfz_mpt=dfz(),dct_mdl={'_id':'','fld':'','clm_x':[],'clm_y':[],'drp':None,'kep':None})
    >>> mdl.drp_dcm_na(mdl.clm_x)       # 删除为空的数据行
    >>> mdl.mdl_nit()                   # 刷新参数集状态, 需要在self.dts发生行列变动时执行
    >>> mdl.mdl_smp_mpt()               # 根据参数集生成数据集 [None, 'smp_trn', 'smp_tst', 'smp_raw']
    >>> mdl.rnd_frs_clf()               # 随机森林变量重要性检验
    >>> mdl.fct_fit(10, prm='chksav')   # 拟合模型 ['chk', 'sav']
    >>> mdl.fct_transform(['电话积极','带看高效','上线稳定','电话稳定','在岗稳定','带看稳定'])  # 应用模型
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        :param dfz_mpt: a dfz for inserted
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param dct_mdl: more params are auto generated, in ['fls', 'smp_raw/_y_0/_y_1/_trn/_tst']
        """
        super(mdz_fct, self).__init__(dts, lcn, spr=spr, dfz_mpt=dfz_mpt, dct_mdl=dct_mdl)

    def fct_fit(self, nfc=3, mtd='principal', rtt='varimax', *, prm='chksav'):
        """
        factor fit step, to check or save model
        :param nfc: number of factors, default 3
        :param mtd: method, default 'principal'
        :param rtt: rotation method, default 'varimax'
        :param prm: in ['chk', 'sav', ''] for [model check, save to file, import self.mdl only]
        :return: None
        """
        self.mdl = FactorAnalyzer(n_factors=nfc, method=mtd, rotation=rtt)
        self.mdl.fit(self._x)
        if re_find('chk', prm):
            self._ct_fit_chk()
        if re_find('sav', prm):
            self.mdl_sav()

    def _ct_fit_chk(self, prm='kmocmmegnvrnlod'):
        """
        几种常见的factor检查方法
        :param prm: ['kmo/cmm/egn/vrn/lod'] for kmo, communalities, eigenvalues, variance, load matrix
        :return: indicators and pictures in prm
        """
        if re_find('kmo', prm):
            print('kmo %.4f; bartlett %.4f.' % (self._ct_fit_chk_kmo()))  # kmo
        if re_find('cmm', prm):
            print('communalities: %s' % self.mdl.get_communalities())  # 公因子方差
        if re_find('egn', prm):
            self._ct_fit_chk_egn()  # egn图
        if re_find('vrn', prm):
            self._ct_fit_chk_vrn()  # 因子代表性
        if re_find('lod', prm):
            self._ct_fit_chk_lod()  # 载荷矩阵重要性和命名

    def _ct_fit_chk_kmo(self):
        """
        kmo检验和巴特利特, 前者要求尽可能大, 后者要求<0.05
        :return: [kmo_model, bartlett]
        """
        kmo_all, kmo_model = calculate_kmo(self._x)
        bartlett = round(calculate_bartlett_sphericity(self._x)[1], 4)
        return kmo_model, bartlett

    def _ct_fit_chk_egn(self):
        """
        eigenvalues to check factor number
        :return: a picture of eigenvalues
        """
        ev, v = self.mdl.get_eigenvalues()  # eigen values  特征值; 通常要求大于1
        plt.scatter(range(1, ev.shape[0] + 1), ev)
        plt.plot(range(1, ev.shape[0] + 1), ev)
        plt.title('Scree Plot')
        plt.xlabel('Factors')
        plt.ylabel('Eigenvalue')
        plt.grid()
        plt.show()

    def _ct_fit_chk_vrn(self):
        """
        variance plot to check factor number
        :return: a picture of variance
        """
        vrn = self.mdl.get_factor_variance()  # 因子的累积贡献度
        plt.scatter(range(1, vrn[2].shape[0] + 1), vrn[2])
        plt.plot(range(1, vrn[2].shape[0] + 1), vrn[2])
        plt.title('Scree Plot')
        plt.xlabel('Factors')
        plt.ylabel('variance')
        plt.grid()
        plt.show()

    def _ct_fit_chk_lod(self, *, prn=False):
        """
        load plot to check factor number
        :param prn: save the picture or not, default False
        :return: a picture of load matrix
        """
        rtr = Rotator()
        load_matrix = rtr.fit_transform(self.mdl.loadings_)
        sns.set(font="simhei")
        df_cm = DataFrame(np_abs(load_matrix), index=self._x.columns)
        plt.figure(figsize=(8, 8))
        ax = sns.heatmap(df_cm, annot=True, cmap="BuPu")
        ax.yaxis.set_tick_params(labelsize=10)  # 设置y轴的字体的大小
        plt.title('Factor Analysis', fontsize='xx-large')
        plt.ylabel('Sepal Width', fontsize='xx-large')  # Set y-axis label
        plt.show()
        if prn:
            plt.savefig('factorAnalysis.png', dpi=300)

    def fct_transform(self, rnm=None, *, rtn=False):
        """
        factor analysis transform by self.mdl, merge factors on the left side of self.dts.
        :param rnm: new name of factors in list
        :param rtn: if return factors or not, default False
        :return: if rtn is True, return factors in type dataframe
        """
        if self.mdl is None:
            self.mdl_lod()
        vrn = self.mdl.get_factor_variance()
        dtf_fct = self.mdl.transform(self._x)
        # 构造命名
        dct_fct = {i: 'fct_' + str(i) for i in range(len(vrn[0]))}
        if rnm is not None:
            dct_tmp = {i: rnm[i] for i in range(len(rnm))} if type(rnm) in [list] else rnm.copy()
            dct_fct.update(dct_tmp)
        # 计算总分
        lst = []
        for i in dtf_fct:
            lst.append(sum([i[j] * vrn[1][j] for j in range(len(i))]))
        dtf_fnl = DataFrame(dtf_fct, index=self._x.index).rename(columns=dct_fct)
        dtf_fnl['scr'] = lst
        # 将结果附加的根源数据集self.dts中
        self.mrg_dtf(dtf_fnl, prm='outer_index')
        if rtn:
            return dtf_fnl


class mdz_kmn(mdz_nit):
    """
    model factor analysis
    >>> mdl = mdz(DataFrame(),dfz_mpt=dfz(),dct_mdl={'_id':'','fld':'','clm_x':[],'clm_y':[],'drp':None,'kep':None})
    >>> mdl.mdl_nit({'_id':''})         # 切换某些dct_mdl中的参数需要使用这个句子
    >>> mdl.mdl_smp_mpt()               # 根据参数集生成数据集 [None, 'smp_trn', 'smp_tst', 'smp_raw']
    >>> mdl.kmn_fit([2,6])              # 测试二到六个聚类的效果
    >>> mdl.kmn_fit(3, prm='sav')       # 经过测试, 确定3格局类最好时对模型进行保存
    >>> mdl.kmn_transform('类型')        # 应用上一步存档的模型, 对分类变量命名为'类型'
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        :param dfz_mpt: a dfz for inserted
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param dct_mdl: more params are auto generated, in ['fls', 'smp_raw/_y_0/_y_1/_trn/_tst']
        """
        super(mdz_kmn, self).__init__(dts, lcn, spr=spr, dfz_mpt=dfz_mpt, dct_mdl=dct_mdl)

    def kmn_fit(self, ncl=None, *, prm='sav'):
        """
        function the best clusters for k-menas.
        :param ncl: iter clusters, default [2,6]
        :param prm: in ['sav'] for save .pkl file to local path
        @return: None
        """
        ncl = lsz(ncl).typ_to_lst(rtn=True) if ncl is not None else [2, 6]
        ncl = range(ncl[0], ncl[1]) if len(ncl) == 2 else ncl  # [None, [2,4], 3] to [[2,6], [2,4], [3]]
        for i in ncl:
            self.mdl = KMeans(n_clusters=i, max_iter=1000)
            self.mdl.fit(self._x)
            y_lbl = self.mdl.labels_
            print('cluster %i; inertia %i; silhouette %.4f; calinski_harabasz %i;' % (
                i,  # cluster number
                self.mdl.inertia_,  # 距离越小说明簇分的越好
                metrics.silhouette_score(self._x, y_lbl, metric='euclidean'),  # 越接近1越好, range[-1, 1]
                metrics.calinski_harabasz_score(self._x, y_lbl)  # 越大越好
            ))
        if prm in ['sav', 'save'] and len(ncl) == 1:  # 只有ncl唯一, 非测试环节时方进行保存
            self.mdl_sav()

    def kmn_transform(self, rnm=None, *, rtn=False):
        """
        K-means transform
        :param rnm: rename in type str, only one new column need this rename param
        :param rtn: return a dataframe or not, default False
        :return: if rtn is True, return a dataframe of cluster label
        """
        rnm = 'clr' if rnm is None else rnm
        if self.mdl is None:
            self.mdl_lod()
        self.mdl.fit(self._x)
        y_lbl = self.mdl.labels_
        self.mrg_dtf(DataFrame(y_lbl, columns=[rnm], index=self._x.index), prm='outer_index')
        if rtn:
            return DataFrame(y_lbl, columns=[rnm], index=self._x.index)


class mdz_tsn(mdz_nit):
    """
    model factor analysis
    >>> mdl = mdz(DataFrame(),dfz_mpt=dfz(),dct_mdl={'_id':'','fld':'','clm_x':[],'clm_y':[],'drp':None,'kep':None})
    >>> mdl.mdl_nit({})
    >>> mdl.mdl_smp_mpt('smp_trn')  # 根据参数集生成数据集 [None, 'smp_trn', 'smp_tst', 'smp_raw']
    >>> mdl.tsn_fit_transform(rnm=['a','b'], prm='savplt')
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        :param dfz_mpt: a dfz for inserted
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param dct_mdl: more params are auto generated, in ['fls', 'smp_raw/_y_0/_y_1/_trn/_tst']
        """
        super(mdz_tsn, self).__init__(dts, lcn, spr=spr, dfz_mpt=dfz_mpt, dct_mdl=dct_mdl)

    def tsn_fit_transform(self, ncl=None, rnm=None, *, prm='savplt', rtn=False, tsn_init='random', tsn_random_state=0):
        """
        t-SNE fit and transform in one step
        :param ncl: number of components(factors)
        :param rnm: rename new factors in type list,
        :param prm: when find 'sav', save local .pkl file; when find 'plt', draw a 2D plot for the result
        :param rtn: if return the components in DataFrame or not, default False
        :param tsn_init: TSNE param
        :param tsn_random_state: TSNE paran
        :return: if rtn is True, return the result in a DataFrame
        """
        ncl = 2 if ncl is None else ncl
        # 构造命名
        dct_rnm = {i: 'tsn_' + str(i) for i in range(ncl)}
        if rnm is not None:
            dct_tmp = {i: rnm[i] for i in range(ncl)} if type(rnm) in [list] else rnm.copy()
            dct_rnm.update(dct_tmp)
        # 建模环节
        self.mdl = TSNE(n_components=ncl, init=tsn_init, random_state=tsn_random_state)
        y_ncl = self.mdl.fit_transform(self._x)
        y_dtf = DataFrame(y_ncl, index=self._x.index).rename(columns=dct_rnm)
        self.mrg_dtf(y_dtf, prm='outer_index')
        if re_find('sav', prm):  # 保存
            self.mdl_sav()
        if re_find('plt', prm) and ncl == 2:  # 当降维2时输出散点图
            self._sn_plt(y_dtf)
        if rtn:
            return y_dtf

    def _sn_plt(self, y_dtf, y_lbl=None):
        """
        t-SNE plot in 2D
        """
        y_ncl = dfz(y_dtf).dtf_to_typ(typ='list', rtn=True, prm='split')['data']
        x_min, x_max = np_min(y_ncl, 0), np_max(y_ncl, 0)
        y_tsn = (y_ncl - x_min) / (x_max - x_min)  # 范围标准化
        y_lbl = DataFrame(self._y) if y_lbl is None else y_lbl  # 标志变量在图中展示, 需要其为序数变量为佳, 可区分颜色
        lst_typ_y = [int, float, np_int32, np_int64, np_float64]
        y_lbl = np_array([[i] for i in np_ravel(y_lbl)]) if type(y_lbl.iloc[0, 0]) in lst_typ_y else y_lbl
        plt.figure()
        for i in range(y_tsn.shape[0]):
            plt.text(
                y_tsn[i, 0], y_tsn[i, 1], (y_lbl[i, 0]),  # y为shape(n,1)的ndarray
                color=plt.cm.Set1(1 - y_lbl[i, 0] / len(list(set([i[0] for i in y_lbl])))),  # y为序列时选择此行
                fontdict={'weight': 'bold', 'size': 9}
            )
            plt.xlabel(y_dtf.columns[0], fontsize=14)
            plt.ylabel(y_dtf.columns[1], fontsize=14)
        plt.show()


class mdz_svm(mdz_nit):
    """
    model factor analysis
    >>> mdl = mdz(DataFrame(),dfz_mpt=dfz(),dct_mdl={'_id':'','fld':'','clm_x':[],'clm_y':[],'drp':None,'kep':None})
    >>> mdl.mdl_nit({})
    >>> mdl.mdl_smp_mpt('smp_trn')  # 建模时使用参数'smp_trn', 应用模型则留空
    >>> mdl.svm_fit(prm='chksav')   # 建模, 展示效果并保存
    >>> mdl.svm_transform(rnm=['prb','rst'])
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        :param dfz_mpt: a dfz for inserted
        :param dct_mdl: a dict of params, keys in ['_id', 'fld', 'clm_y', 'clm_x', 'drp', 'kep']
        :param dct_mdl: more params are auto generated, in ['fls', 'smp_raw/_y_0/_y_1/_trn/_tst']
        """
        super(mdz_svm, self).__init__(dts, lcn, spr=spr, dfz_mpt=dfz_mpt, dct_mdl=dct_mdl)

    def svm_fit(self, typ='SVC', *, prm='chksav', svm_kernel='rbf'):
        """
        SVM fit step
        :param typ: in ['SVC', 'SVR'] for classifier and regression
        :param prm: check confusion matrix and precision when find 'chk', save to .pkl file when find 'sav'
        :param svm_kernel: SVM param
        :return: None
        """
        if typ in ['SVC', 'svc']:       # 分类器
            self.mdl = SVC(kernel=svm_kernel, probability=True)
        elif typ in ['SVR', 'svr']:     # 回归器
            self.mdl = SVR(kernel=svm_kernel)
        self.mdl.fit(self._x, np_ravel(self._y))
        if re_find('chk', prm) and type(self.mdl) in typ_svc:
            self._vm_fit_chk()
        if re_find('sav', prm):
            self.mdl_sav()

    def _vm_fit_chk(self, lst_sch=None):
        y_tst, x_tst = self.mdl_smp_mpt('smp_tst', rtn=True)  # 对测试集进行验证
        lst_sch = [0.4, 0.7, 20] if lst_sch is None else lst_sch  # from [0] to [1], batches number [2]
        for i in range(lst_sch[2]):
            y_bln = lst_sch[0] + i * (lst_sch[1] - lst_sch[0]) / lst_sch[2]
            y_pre = [1 if i[1] >= y_bln else 0 for i in self.mdl.predict_proba(x_tst)]  # y_pre = mdl_svm.predict(x_tst)
            cfs = metrics.confusion_matrix(y_tst, y_pre)
            cfs = DataFrame(cfs, columns=['p', 'n'], index=['T', 'F'])
            print('confusionMatrix and precision in %.2f:' % y_bln)
            print(cfs, '; ', round((cfs.iloc[1, 1] + cfs.iloc[0, 0]) / (cfs.values.sum()), 2), '; ')

    def svm_transform(self, rnm=None, prb=0.5, *, rtn=False):
        """
        SVM transform step
        :param rnm: rename [prob, result], need a list in length 2
        :param prb: 概率分割线, 当prob大于这个值时判断class为1, 否则为0
        :param rtn: return the result in DataFrame or not
        :return: if rtn is True, return the target DataFrame
        """
        if self.mdl is None:
            self.mdl_lod()
        # 构造命名
        dct_rnm = {0: self.lcn['mdl']['clm_y'][0] + '_prb', 1: self.lcn['mdl']['clm_y'][0] + '_rst'}
        if rnm is not None:
            dct_tmp = {i: rnm[i] for i in range(len(rnm))} if type(rnm) in [list] else rnm.copy()
            dct_rnm.update(dct_tmp)
        # 效果判断
        y_prb = []
        if type(self.mdl) in typ_svc:
            y_prb = [[i[1], 1 if i[1] >= prb else 0] for i in self.mdl.predict_proba(self._x)]
        elif type(self.mdl) in typ_svr:
            y_prb = [[i, 1 if i >= prb else 0] for i in self.mdl.predict(self._x)]
        y_prb = DataFrame(y_prb, index=self._x.index).rename(columns=dct_rnm)
        self.mrg_dtf(y_prb, prm='outer_index')
        if rtn:
            return y_prb


class mdz(mdz_fct, mdz_kmn, mdz_tsn, mdz_svm):
    """
    model class mixin
    """
    def __init__(self, dts=None, lcn=None, *, spr=False, dfz_mpt=None, dct_mdl=None):
        super(mdz, self).__init__(dts, lcn, spr=spr, dfz_mpt=dfz_mpt, dct_mdl=dct_mdl)

    def rnd_frs_clf(self, x=None, y=None, *, estimator=100):
        """random Forest clarify for importance of columns"""
        x = self._x if x is None else x
        y = self._y if y is None else y
        frs = RandomForestClassifier(n_estimators=estimator, random_state=0, n_jobs=-1)
        frs.fit(x, np_ravel(y))
        x_importance = frs.feature_importances_
        x_name = x.columns[0:]
        indices = np_argsort(x_importance)[::-1]
        for f in range(x.shape[1]):
            print("%2d) %-*s %.4f" % (f + 1, 30, x_name[indices[f]], x_importance[indices[f]]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.18
setup for package.
@author: zoharslong
"""
from setuptools import setup, find_packages
from os.path import join as os_join, abspath as os_abspath, dirname as os_dirname

here = os_abspath(os_dirname(__file__))
with open(os_join(here, 'README.md')) as f:
    README = f.read()

setup(
    name="pyzohar",
    version="0.1.12",
    author="zoharslong",
    author_email="zoharslong@hotmail.com",
    description="a private package on data pre-processing.",
    long_description=README,
    url="https://www.xzzsmeadow.com/",
    license="MIT",
    classifiers=[
        'Development Status :: 3 - Alpha',  # {3:Alpha, 4:Beta, 5:Production/Stable}
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(),
    keywords='data pre-processing',
    python_requires='>=3',
    install_requires=[
        'numpy>=1.18.1',
        'pandas>=1.0.1',
        'pymongo>=3.9.0',
        'pymysql>=0.9.3',
        'fake-useragent>=0.1.11',
        'requests>=2.22.0',
        'openpyxl>=3.0.3',  # excel files resolving
        'urllib3>=1.25.8',  # some error type of http requests
        # 'matplotlib>=3.1.3',  # for sub_slt_mdl.mdz
        # 'sklearn>=0.22.1',  # for sub_slt_mdl.mdz
        # 'seaborn>=0.10.0',  # for sub_slt_mdl.mdz
        # 'factor_analyzer>=0.3.2',  # for sub_slt_mdl.mdz
        # 'joblib>=0.14.1',  # for sub_slt_mdl.mdz
        # 'python-pptx>=0.6.19',  # for sub_slt_ppt.ppz
    ],
    package_data={'pyzohar': ['samples/*.*']},
    include_package_data=True,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
A package slot for web spider
@author: zoharslong
"""
from pyzohar.sub_slt_spd.spz import spz

__all__ = ['spz']
print('info: web spider slot inserted.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 16:30:38 2019
dataframe operation.
@author: zoharslong
"""
from numpy import nan as np_nan, max as np_max, min as np_min, sum as np_sum, std as np_std
from numpy import mean as np_mean
from pandas import merge, concat, DataFrame, cut, isnull as pd_isnull
from re import search as re_search, findall as re_findall, sub as re_sub, match as re_match
from pandas.core.indexes.base import Index as typ_pd_Index              # 定义dataframe.columns类型
from math import isnan as math_isnan
from json import loads
from pyzohar.sub_slt_bsc.bsz import stz, lsz, dtz
from pyzohar.sub_slt_bsc.ioz import ioz


class dfBsc(ioz):
    """
    dataFrame Basic.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(dfBsc, self).__init__(dts, lcn, spr=spr)


class clmMixin(dfBsc):
    """
    methods Mixin for columns.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(clmMixin, self).__init__(dts, lcn, spr=spr)

    def srt_clm(self, lst_clm, *, spr=False, rtn=False, drp=True):
        """
        sort columns.
        >>> from pandas import DataFrame
        >>> tst = clmMixin(DataFrame([{'A':1,'B':1}]),spr=True)
        >>> tst.srt_clm('A', rtn=True)
           A
        0  1
        :param lst_clm: columns to be sorted
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param drp: drop columns which not mentioned in lst_clm, default True
        :return: if rtn is Ture, return self.dts
        """
        lst_fnl = []
        lst_clm = lsz(lst_clm).typ_to_lst(rtn=True)
        for i_cll in lst_clm:  # 对self.dts中存在的columns, 按照lst_clm_srt排序
            if list({i_cll} & set(self.clm)):
                lst_fnl.append(i_cll)
        if drp:  # bln_drp为True，则舍弃self.clm中，未出现再lst_clm_srt中的列
            pass
        else:  # bln_drp为False，则将self.clm中有而lst_clm_srt未提及的列附加在后面
            x = lsz(lst_fnl).mrg('differ', self.clm, rtn=True)
            lst_fnl.extend(x)
        self.dts = self.dts[lst_fnl]
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_clm(self, lst_clm, *, spr=False, rtn=False, prn=True):
        """
        drop columns.
        >>> from pandas import DataFrame
        >>> tst = clmMixin(DataFrame([{'A':1,'B':1}]),spr=True)
        >>> tst.drp_clm('A', rtn=True)
           B
        0  1
        :param lst_clm: a list of columns to drop
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prn: print info or not
        :return: if rtn is True, return self.dts
        """
        lst_clm = lsz(lst_clm)
        lst_clm.typ_to_lst(rtn=True)
        lst_clm = lst_clm.mrg('inter', list(self.clm), rtn=True)
        self.dts = self.dts.drop(lst_clm, axis=1)
        if prn:
            print('info: found columns %s in target dataFrame, dropping.' % str(lst_clm))
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def rnm_clm(self, *args, spr=False, rtn=False):
        """
        rename columns.
        >>> from pandas import DataFrame
        >>> tst = clmMixin(DataFrame([{'A':1,'B':1}]),spr=True)
        >>> tst.rnm_clm({'A':'a'}, rtn=True) # tst.rnm_clm('A','a',rtn=True)
           a  B
        0  1  1
        :param args: old columns' name and the new. 填入两个列表或一个字典，表达重命名对应关系
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn is True, return self.dts, default False
        """
        if len(args) == 1 and type(args[0]) in [dict]:
            self.dts = self.dts.rename(columns=args[0])
        else:
            self.dts = self.dts.rename(columns=lsz(args[1]).lst_to_typ('dct', args[0], rtn=True)[0])
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def fll_clm_na(self, *args, spr=False, rtn=False, prm=None):
        """
        fill columns' na cells.
        >>> tst = clmMixin(DataFrame([{'A':np_nan,'B':1}]),spr=True)
        >>> tst.fll_clm_na({'A':1}, rtn=True)
             A  B
        0  1.0  1
        >>> tst = dfz(DataFrame([{'A':1,'B':1},{'A':np_nan,'B':1}]),spr=True)
        >>> tst.fll_clm_na(prm='ffill')
             A  B
        0  1.0  1
        1  1.0  1
        :param args: columns and values to be filled. 填入两个列表或一个字典，表达对列中空缺值的填充关系
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: method of fill na, in ['backfill', 'bfill', 'pad', 'ffill', None]
        :return: if rtn is True, return self.dts
        """
        if len(args) == 0:
            dct_fll = None
        elif len(args) == 1:
            dct_fll = args[0]
        else:
            dct_fll = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        self.dts = self.dts.fillna(value=dct_fll, method=prm)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_clm_na(self, *, spr=False, rtn=False, prm=None):
        """
        drop columns for nan.
        >>> from pandas import DataFrame
        >>> tst = clmMixin(DataFrame([{"A":np_nan,'B':1}]))
        >>> tst.drp_clm_na(rtn=True)
           B
        0  1
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: if na numbers in one columns bigger than prm, delete the columns
        :return: if rtn is True, return self.dts
        """
        self.dts = self.dts.dropna(axis=1,          # 0 for index and 1 for columns, if 0, subset is available
                                   how='any',       # any and all
                                   thresh=prm,      # optional drop the column if it has at least thresh nan
                                   subset=None,)    # optional columns in target
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def stt_clm(self, lst_clm, *args, spr=False, rtn=False, prm=None):
        """
        static columns.
        >>> from pandas import DataFrame
        >>> tst = clmMixin(DataFrame([{'A':1,'B':2},{'A':1,'B':3},{'A':2,'B':4},{'A':2,'B':5},]))
        >>> tst.stt_clm(['A'], ['B'], [np_sum], rtn=True, prm='C')  # tst.stt_clm(['A'],{'B':np_sum},rtn=True,prm='C')
           A  C
        0  1  5
        1  2  9
        :param lst_clm: group by columns
        :param args: static columns and methods, in format ({columns:method}) or ([columns], [method])
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: new name for static columns, default remain old names if None
        :return: if rtn is True, return self.dts
        """
        dct_stt = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_stt = list(dct_stt.keys())
        self.dts = self.dts.groupby(lsz(lst_clm).typ_to_lst(rtn=True))
        self.dts = self.dts.aggregate(dct_stt).reset_index(drop=False)
        if type(self.clm) is typ_pd_Index:
            lst_rnm = lsz(prm).typ_to_lst(rtn=True) if prm else lst_stt
            self.rnm_clm(lst_stt, lst_rnm)
        else:
            print("info: columns' multi-index cannot be altered.")
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_rgx(self, *args, spr=False, rtn=False, prm=None):
        """
        add new columns from old columns' some part.
        >>> tst = dfz([{"A":'1=a;'},{"A":'2=b;'}])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_rgx('A', 'B', prm='1=(.*?);', rtn=True)
              A     B
        0  1=a;     a
        1  2=b;  None
        :param args: ({oldColumns: newColumns}) or ([oldColumns], [newColumns])
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: a list of regex to be found in old columns
        :return: if rtn is True, return self.dts
        """
        lst_rgs = lsz(args).rgs_to_typ(prm='eql', rtn=True)
        lst_clm, lst_new = lsz(lst_rgs[0]).typ_to_lst(rtn=True), lsz(lst_rgs[1]).typ_to_lst(rtn=True)
        prm = lsz(prm)
        prm.typ_to_lst()
        prm.cpy_tal(len(lst_clm), spr=True)
        for i in range(len(lst_clm)):
            self.dts[lst_new[i]] = [re_findall(prm[i], j)[0] if type(j) in [str] and re_search(prm[i], j) else None for
                                    j in self.dts[lst_clm[i]]]
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_dtf(self, str_clm, *, spr=False, rtn=False):
        """
        若dataframe中某列中为形如[{},{},..]的可转化为dataframe的结构，则解出拼接至原dataframe，导致原dataframe的行列均扩增
        :param str_clm: 列单元格内容为[{},{},..]格式的列名
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: None
        """
        dtf_dtl = DataFrame([])
        for i in range(self.len):
            self.dts[str_clm][i] = loads(self.dts[str_clm][i].replace("'", '"')) if \
                type(self.dts[str_clm][i]) in [str] else self.dts[str_clm][i]
            len_tmp = len(self.dts[str_clm][i]) if type(self.dts[str_clm][i]) is list else 1  # 兼容{}, [{},{}]两种输入
            dtf_i = DataFrame(self.dts[str_clm][i], index=[int(i)]*len_tmp)
            dtf_dtl = concat([dtf_dtl, dtf_i])  # default prm axis=0, sort=False
        self.dts.drop([str_clm], axis=1, inplace=True)
        self.dts = concat([dtf_dtl, self.dts], axis=1)  # default prm join='outer', sort=False
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_spl(self, clm_spl, lst_vls, clm_rmn=None, *, spr=False, rtn=False):
        """
        add columns from spliting an column.
        >>> slf = dfz([{'est':'A','date':'1','vls':1},{'est':'B','date':'1','vls':2},
        >>>             {'est':'A','date':'2','vls':1},{'est':'B','date':'3','vls':2},])
        >>> slf.typ_to_dtf()
        >>> slf.add_clm_spl('est','vls', rtn=True)  # 按照est列，将vls列的值扩张成多个新的列
          date  vls_A  vls_B
        0    1    1.0    2.0
        1    3    1.0    NaN
        2    2    NaN    2.0
        :param clm_spl: 用于分出多列的列，列中的每一种唯一元素都会成为新的列
        :param lst_vls: 用于分到多列的值
        :param clm_rmn: 保留到新表中的用作拼接的列
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn, return self.dts
        """
        dtf_fnl = DataFrame([])
        clm_spl_cll = self.dts[clm_spl].unique()
        lst_vls = lsz(lst_vls).typ_to_lst(rtn=True)
        # 对clm_rmn进行预处理
        if clm_rmn is None:
            clm_ttl = [i for i in lsz(self.clm.copy()).typ_to_lst(rtn=True) if i not in [clm_spl]]
            clm_rmn = [i for i in clm_ttl if i not in lst_vls+[clm_spl]]
        else:
            clm_rmn = [i for i in lsz(clm_rmn).typ_to_lst(rtn=True) if i not in lst_vls+[clm_spl]]
            clm_ttl = lsz(clm_rmn + lst_vls).typ_to_lst(rtn=True)
        print(clm_ttl)
        # 如果存在标志列重复的行，则终止本次列转行操作，因为会在拼接过程中导致虚增重复的行数
        len_bfr = self.len
        chk_dpl = self.dts.drop_duplicates(subset=[clm_spl] + clm_rmn, keep='first')
        if len_bfr != chk_dpl.shape[0]:
            raise AttributeError("stop: duplicate documents exist, %.i / %.i." % (chk_dpl.shape[0], len_bfr))
        # clm_spl中每个唯一的cell进行一次clm_vls列的值转新列的操作
        for i in range(clm_spl_cll.shape[0]):
            dtf_tmp = self.dts.loc[self.dts[clm_spl] == clm_spl_cll[i], :][clm_ttl]
            dtf_tmp.rename(columns={j: j+'_'+clm_spl_cll[i] for j in lst_vls}, inplace=True)
            dtf_fnl = dtf_tmp.copy() if dtf_fnl.shape[0] == 0 else merge(dtf_fnl, dtf_tmp, how='outer', on=clm_rmn)
        self.dts = dtf_fnl
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_ctg(self, *args, prm='num', ctg_num=2, ctg_edg=None, ctg_lbl=None, bln_lwr=False, spr=False, rtn=False):
        """
        add new columns of certain columns in category.
        >>> tst = dfz([{'A':1},{'A':2},{'A':3},{'A':4},{'A':5},])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_ctg('A',ctg_lbl=['a','b','c','d','e'],bln_lwr=True)  # 按默认参数自动按值分成两箱
           A
        0  a
        1  a
        2  a
        3  b
        4  b
        >>> tst = dfz([{'A':1},{'A':2},{'A':3},{'A':4},{'A':5},])
        >>> tst.typ_to_dtf()
        >>> lst_sqr_cut = [0,40,60,80,100,120,140,160]
        >>> lst_sqr_lbl = ['-40','40-60','60-80','80-100','100-120','120-140','140-']
        >>> tst.add_clm_ctg('A', 'a', prm=lst_sqr_cut, ctg_lbl=lst_sqr_lbl)  # 指定分箱界限
        :param args: {'targetColumn':'newColumn'}
        :param prm: in ['number', 'num', 'width']
        :param ctg_num: the number of boxes if prm in ['num', 'number'], the width of each box if prm in ['width']
        :param ctg_edg: the min and max limitation of th category
        :param ctg_lbl: label name of each boxes
        :param bln_lwr: method in (a, b] or [a, b], default (,]
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn, return self.dts
        """
        lst_rgs = lsz(args).rgs_to_typ(rtn=True)
        str_mpt, str_xpt = lst_rgs[0], lst_rgs[1]
        flt_min = np_min(self.dts[str_mpt])     # 自动产生分箱的上下界
        flt_max = np_max(self.dts[str_mpt])
        if ctg_edg is not None and ctg_edg[0] <= flt_min and ctg_edg[1] >= flt_max:
            flt_min = ctg_edg[0]
            flt_max = ctg_edg[1]
        elif ctg_edg is None:
            pass
        else:
            raise AttributeError('stop: edge of the category is not available.')
        # 构建分箱规则列表
        flt_cts = (flt_max - flt_min) / ctg_num
        if prm in ['num']:        # option: 确定分箱数量
            lst_bns = [flt_min] + [flt_min+(i+1)*flt_cts for i in range(int(ctg_num))]
        elif prm in ['width']:    # option: 确定分箱宽度
            from math import ceil
            lst_bns = [flt_min] + [flt_min+(i+1)*ctg_num for i in range(ceil(flt_cts))]
        else:
            lst_bns = lsz(prm).typ_to_lst(rtn=True)  # 也可以在prm中直接指定分箱分界线
        # 基础分箱
        ctg_lbl = ctg_lbl[:len(lst_bns) - 1] if ctg_lbl is not None else ctg_lbl
        self.dts[str_xpt] = cut(self.dts[str_mpt], lst_bns, labels=ctg_lbl, include_lowest=bln_lwr)
        self.dts_nit()
        print('info: transform column <%s> in category %s.' % (str_mpt, str(ctg_lbl)))
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def ltr_clm_rpc(self, lst_clm, *args, spr=False, rtn=False, prm=None):
        """
        alter columns by replacing contents.
        >>> tst = clmMixin([{'A':'1'},{'A':'abc'},{'A':None}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_rpc('A','1','2', rtn=True)
              A
        0     2
        1   abc
        2  None
        >>> tst.ltr_clm_rpc('A','b','B',prm='part', rtn=True)
              A
        0     2
        1   aBc
        2  None
        >>> from math import isnan
        >>> tst = clmMixin([{'A':1},{'A':None}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_rpc('A',{'':'thisIsNa'},prm=isnan, rtn=True)
                  A
        0         1
        1  thisIsNa
        :param lst_clm: target columns, all the alteration are used on each column
        :param args: ({old_content:new_content}) or (old_content, new_content)
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: in [None, 'part', function]
        :return: if rtn is True, return self.dts
        """
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_old, lst_new = list(dct_rgs.keys()), list(dct_rgs.values())
        lst_clm = lsz(lst_clm).typ_to_lst(rtn=True)
        for i in range(len(lst_clm)):
            for j in range(len(lst_old)):
                if prm is None:                     # 当prm为None时运行完全替换
                    self.dts[lst_clm[i]] = [lst_new[j] if k == lst_old[j] else k for k in self.dts[lst_clm[i]]]
                elif prm in ['part', 'prt']:    # 当prm为string类型且内容为'part'时运行部分替换
                    self.dts[lst_clm[i]] = [re_sub(lst_old[j], lst_new[j], k) if
                                            type(k) is str and re_search(lst_old[j], k) else
                                            k for k in self.dts[lst_clm[i]]]
                else:                           # 当prm为函数时判断满足prm()则运行替换
                    self.dts[lst_clm[i]] = [lst_new[j] if prm(k) else k for k in self.dts[lst_clm[i]]]
        self.dts_nit(False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def ltr_clm_unl(self, *args, spr=False, rtn=False):
        """
        atler columns in type str English words to upper or lower.
        >>> tst = clmMixin([{'A':1},{'A':'abC'},{'A':'ABc'}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_unl('A','ppr', rtn=True)
             A
        0    1
        1  ABC
        2  ABC
        :param args: ({columns:upperNLower}) or ([columns],[upperNLower]), unl in ['upper','ppr','lower','lwr']
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn is True, return self.dts
        """
        # *args预处理
        dct_typ = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        # 转化部分
        for i_clm in [i for i in dct_typ.keys() if dct_typ[i] in ['lower', 'lwr']]:
            self.dts[i_clm] = self.dts.apply(lambda x: x[i_clm].lower() if type(x[i_clm]) is str else x[i_clm], axis=1)
        for i_clm in [i for i in dct_typ.keys() if dct_typ[i] in ['upper', 'ppr']]:
            self.dts[i_clm] = self.dts.apply(lambda x: x[i_clm].upper() if type(x[i_clm]) is str else x[i_clm], axis=1)
        self.dts_nit(False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def ltr_clm_flt(self, *args, spr=False, rtn=False, prn=False):
        """
        check columns in float, if cannot convert into float, fill None.
        >>> tst = clmMixin([{'A':1.21},{'A':2},{'A':'a'}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_flt('A')
        [{'A': ['a']}]  # 返回格式[{column: [unusual cells, ...]},{},...]
        :param args: target columns, alter if len == 1, create new columns if len == 2
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prn: print cells that cannot convert into float
        :return: if prn is True, return unusual cells; if not prn and rtn, return self.dts
        """
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_old, lst_clm = list(dct_rgs.keys()), list(dct_rgs.values())
        lst_ttl = []
        for i in range(len(lst_clm)):
            if lst_old[i] in self.clm:
                lst_err = [i for i in self.dts[lst_old[i]] if
                           (not re_match("^\d+?$", str(i))) & (not re_match("^\d+?\.\d+?$", str(i)))]
                self.dts[lst_clm[i]] = self.dts.apply(lambda x: None if
                                                      (not re_match("^\d+?$", str(x[lst_old[i]]))) &
                                                      (not re_match("^\d+?\.\d+?$", str(x[lst_old[i]]))) else
                                                      str(x[lst_old[i]]), axis=1)
                lst_ttl.append({lst_old[i]: lst_err})
        if prn:
            print('info: unusual cells %s.' % lst_ttl)
            return lst_ttl
        self.dts_nit(False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def ltr_clm_rnd(self, *args, spr=False, rtn=False, prm=2):
        """
        alter columns' decimal scales.
        >>> tst = dfz([{'A':1.111},{'A':1.1},{'A':None}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_rnd('A',prm=1, rtn=True)
             A
        0  1.1
        1  1.1
        2  NaN
        :param args: ([oldColumns]) or ([oldColumns],[newColumns])
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: decimal scales. default 2
        :return: if rtn is True, return self.dts
        """
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_old, lst_clm = list(dct_rgs.keys()), list(dct_rgs.values())
        prm = lsz(prm)
        prm.typ_to_lst()
        prm.cpy_tal(len(lst_clm), spr=True)
        for i in range(len(lst_clm)):
            if prm[i] != 0 and lst_old[i] in self.clm:
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: np_nan if pd_isnull(x[lst_old[i]]) else round(float(x[lst_old[i]]), prm[i]), axis=1
                )
            elif lst_old[i] in self.clm:
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: np_nan if pd_isnull(x[lst_old[i]]) else int(x[lst_old[i]]), axis=1
                )
            else:
                print('info: %s not in the dfz' % lst_old[i])
        self.dts_nit(False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def ltr_clm_typ(self, *args, spr=False, rtn=False, prm=False):
        """
        alter columns' data type.
        >>> tst = clmMixin([{'A':1,'B':'a'},{'A':2,'B': 3},{'A':'', 'B':None}])
        >>> tst.typ_to_dtf()
        >>> tst.ltr_clm_typ({'A':float,'B':str}, rtn=True)  # 空类值在转float时会转为np.nan, 转str时会转为''
        >>> # 空值类形如['NA', 'nan', 'null', 'None', '', None, np.nan]
             A  B
        0  1.0  a
        1  2.0  3
        2  NaN
        >>> tst.ltr_clm_typ({'C':float}, prm=True) # 若不确定args column name是否存在仍要求不报错，则设定prm=True
        :param args: ({columns: types}) or ([columns],[types]), types in [str, float, int]
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: check if target column names is in self.clm or not, default False means not check
        :return: if rtn is True, return self.dts
        """
        dct_typ = lsz(args).rgs_to_typ(prm='dct', rtn=True)                # args转字典
        if prm:     # 当prm==True时, 自动排除*args[0]中不存在于self.clm中的列名以避免报错
            for j in [i for i in dct_typ.keys() if i not in self.clm]:
                dct_typ.pop(j)
        lst_flt = [i for i in dct_typ.keys() if dct_typ[i] in [int, 'int', float, 'float']]
        self.ltr_clm_rpc(lst_flt, ['NA', 'N/A', 'nan', 'null', 'None', None, ''], [np_nan])    # 转float列的空值处理
        self.dts = self.dts.astype(dct_typ)  # 核心转换
        lst_str = [i for i in dct_typ.keys() if dct_typ[i] in [str, 'str']]
        self.ltr_clm_rpc(lst_str, ['NA', 'N/A', 'nan', 'null', 'None', None], [''])            # 转str列的空值处理
        self.dts_nit(False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def cod_clm_int(self, lst_clm, *, spr=False, rtn=False, prm='encode'):
        """
        encode or decode int columns with special coding dict.
        :param lst_clm:
        :param spr:
        :param rtn:
        :param prm: in ['encode','decode'], default 'encode'
        :return:
        """
        lst_clm = lsz(lst_clm).typ_to_lst(rtn=True)
        if prm in ['encode', 'ncd']:
            self.ltr_clm_typ(lst_clm, 'str')
            self.ltr_clm_rpc(lst_clm, {'\.\d+': ''}, prm='prt')
            for i in lst_clm:
                self.dts[i] = self.dts.apply(lambda x: stz(x[i]).ncd_int(rtn=True), axis=1)
        elif prm in ['decode', 'dcd']:
            for i in lst_clm:
                self.dts[i] = self.dts.apply(lambda x: stz(x[i]).dcd_int(rtn=True), axis=1)
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def cod_clm_cvr(self, lst_clm, *, spr=False, rtn=False, prm=None):
        """
        encode columns contents with covering by '*'.
        :param lst_clm:
        :param spr:
        :param rtn:
        :param prm:
        :return:
        """
        prm = [1, -1] if prm is None else prm
        lst_clm = lsz(lst_clm).typ_to_lst(rtn=True)
        self.ltr_clm_typ(lst_clm, 'str')
        for i in lst_clm:
            self.dts[i] = self.dts.apply(lambda x: stz(x[i]).ncd_cvr(prm), axis=1)
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def cod_clm_md5(self, *args, spr=False, rtn=False, prm=None):
        """
        endocd column contents by md5.
        :param args: columns' name in a list, format in (targetColumn,), ([targetColumns]), (targetColumn, newColumn)
        :param spr:
        :param rtn:
        :param prm:
        :return:
        """
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_old, lst_clm = list(dct_rgs.keys()), list(dct_rgs.values())
        self.ltr_clm_typ(lst_old, 'str')
        for i in range(len(lst_clm)):
            self.dts[lst_clm[i]] = self.dts.apply(lambda x: stz(x[lst_old[i]]).ncd_md5(rtn=True, prm=prm), axis=1)
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts


class dcmMixin(dfBsc):
    """
    documents' Mix in
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(dcmMixin, self).__init__(dts, lcn, spr=spr)

    def srt_dcm(self, lst_clm=None, scd=True, *, spr=False, rtn=False, prm='last'):
        """
        sort documents.
        :param lst_clm: target columns
        :param scd: in [True, false] for [ascending, descending], default ascending
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: in ['first','last'], default 'last'
        :return: if rtn is True, return self.dts
        """
        self.dts = self.dts.sort_values(by=lst_clm, ascending=scd, na_position=prm)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_dcm(self, lst_ndx, *, spr=False, rtn=False, prm='drop', ndx_rst=True):
        """
        drop documents by index numbers.
        :param lst_ndx: a list of index
        :param prm: in ['drop','keep'], default 'drop'
        :param ndx_rst: 是否重置self.dts的index, default True
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn is True, return self.dts
        """
        if lst_ndx is None:
            pass
        else:
            lst_ndx = lsz(lst_ndx).typ_to_lst(rtn=True)
            if prm in ['keep', 'kep', 'k']:  # 当方法选择为keep时, 保留指定index的行
                lst_ndx = lsz().mrg('differ', list(self.dts.index), lst_ndx, rtn=True)
            self.set_dts(self.dts.drop(lst_ndx, axis=0), ndx_rst=ndx_rst)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_dcm_na(self, lst_clm=None, *, spr=False, rtn=False, prm=None, how='any'):
        """
        drop documents for na cells.
        >>> import pandas as pd
        >>> tst = dfz([{'A':1,'B':'a','C':'a'},{'A':pd.NaT,'B':'b','C':'a'},{'A':pd.NaT,'B':pd.NaT,'C':'a'}])
        >>> tst.typ_to_dtf()
        >>> tst.drp_dcm_na(prm=2, rtn=True)     # 当行中至少有两单元格非空时保留
               A  B  C
        0      1  a  a
        1    NaT  b  a
        >>> tst.drp_dcm_na('B', rtn=True)
               A  B  C
        0      1  a  a
        1    NaT  b  a
        >>> tst.drp_dm_na('A', prm='keep', rtn=True)
               A  B  C
        0    NaT  b  a
        :param lst_clm: default None means do nothing
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm:  remain documents at least <prm> cells not nan, default None
        :param how: in ['any', 'all'] for [任意列存在nan则删除或保留, 全部指定列为nan才删除或保留]
        :return: if rtn is True, return self.dts
        """
        flt_trs = None if prm in ['keep', 'kep', 'kp'] else prm
        flt_bgn = self.len
        lst_clm = lsz(lst_clm).typ_to_lst(rtn=True) if prm in ['keep', 'kep', 'kp', None] else lst_clm
        dts_drp = self.dts.dropna(axis=0,           # 0 for index and 1 for columns
                                  how=how,          # any and all
                                  thresh=flt_trs,   # optional Keep only the rows with at least 2 non-NA values
                                  subset=lst_clm)   # optional columns in target
        if prm in ['keep', 'kep', 'kp']:            # 当prm为'保留'时，反向删除不为空的行
            self.drp_dcm(list(dts_drp.index))
            print('info: %i remains from %i by keeping na.' % (self.len, flt_bgn))
        else:                                       # 当prm未经过特别定义时，对dfz.dts沿用去空操作
            self.dts = dts_drp.copy()
            print('info: %i remains from %i by dropping na.' % (self.len, flt_bgn))
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_dcm_dpl(self, *args, spr=False, rtn=False, prm='first', scd=True, prn=True):
        """
        drop duplicate documents.
        :param args: ([columnsDropping]) or ([columnsDropping], [columnsSorting])
        :param scd: if len(args) == 2, define the sorting method
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: dropping method, in ['first','last', 'keep'], default first, keep means keep duplicates in self.dts
        :param prn: print detail of dropping duplicates or not, default True
        :return: if rtn is True, return self.dts
        """
        lst_dpl = lsz(args[0]).typ_to_lst(rtn=True)
        if len(args) > 1:
            lst_srt = lsz(args[1]).typ_to_lst(rtn=True)
            self.srt_dcm(lst_srt, scd)
        len_bfr = self.len
        if prm.lower() in ['keep', 'kep']:
            self.drp_dcm(list(self.dts.drop_duplicates(subset=lst_dpl, keep=False).index))
        else:
            self.dts = self.dts.drop_duplicates(subset=lst_dpl, keep=prm)
        if prn:
            print("info: drop duplicate documents, %.i in total %.i left, from <%s>." % (len_bfr, self.len, self))
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def drp_dcm_ctt(self, *args, spr=False, rtn=False, prm='drop', ndx_rst=True):
        """
        drop documents for contents.
        >>> tst = dfz([{'A':'a','B':'1'},{'A':'b','B':'2'},{'A':'c','B':'3'},{'A':'d','B':'4'},])
        >>> tst.typ_to_dtf()
        >>> tst.drp_dcm_ctt(['A','B'],[['a','b'],['2','3']],rtn=True)
           A  B
        0  d  4
        :param args: ({columns:[content,..]}) or ([column,..],[[content,..],..])
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: in [(True,'drop'),(False,'keep'),'partDrop','partKeep'], default True for dropping documents
        :param ndx_rst: 是否重置self.dts的index, default True
        :return: if rtn is True, return self.dts
        """
        lst_rgs = lsz(args).rgs_to_typ(rtn=True)
        lst_clm, lst_ctt = lsz(lst_rgs[0], lst=True), lsz(lst_rgs[1], lst=True)
        if lst_clm.len == 1 and type(lst_ctt[0]) not in [list, lsz]:
            lst_ctt = lsz([lst_ctt.seq])                    # 当仅检查单列时，需要控制其待识别内容为唯一个列表
        lst_clm.cpy_tal(lsz([lst_clm.seq, lst_ctt.seq]).len_max, spr=True)
        lst_ctt.cpy_tal(lsz([lst_clm.seq, lst_ctt.seq]).len_max, spr=True)
        for i in range(len(lst_clm)):
            if prm is False or prm in ['keep', 'kep']:      # 当not drop 或prm='keep'时，保留完全匹配的行
                self.set_dts(self.dts.loc[self.dts[lst_clm[i]].isin(lsz(lst_ctt[i]).typ_to_lst(rtn=True)), :],
                             ndx_rst=ndx_rst)
            elif prm is True or prm in ['drop', 'drp']:     # 当drop 或prm='drop'时，删除完全匹配的行
                self.set_dts(self.dts.loc[~self.dts[lst_clm[i]].isin(lsz(lst_ctt[i]).typ_to_lst(rtn=True)), :],
                             ndx_rst=ndx_rst)
            elif prm.lower() in ['partdrop', 'droppart', 'prtdrp', 'drpprt']:   # 当prm='partDrop'时，删除正则匹配的行
                lst_ndx = [k for j in lst_ctt[i] for k in self.dts.index if
                           re_search(j, str(self.dts.loc[k, lst_clm[i]]))]
                self.set_dts(self.dts.drop(lst_ndx, axis=0), ndx_rst=ndx_rst)
            elif prm.lower() in ['partkeep', 'keeppart', 'prtkep', 'kepprt']:   # 当prm='partKeep'时，保留正则匹配的行
                lst_ndx = [k for j in lst_ctt[i] for k in self.dts.index if
                           re_search(j, str(self.dts.loc[k, lst_clm[i]]))]
                lst_ndx = [i for i in self.dts.index if i not in lst_ndx]
                self.set_dts(self.dts.drop(lst_ndx, axis=0), ndx_rst=ndx_rst)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_dcm_spl(self, *args, prm=',', spr=False, rtn=False):
        """
        add document like transpose in SAS, from one column splitting. 利用某个分隔符进行分列并转变为相同结构的多行.
        >>> tst = dfz([{'cls': '1', 'vls':'1,2,3'},
        >>>            {'cls': '2', 'vls':'a,b,c'},])
        >>> tst.typ_to_dtf()
        >>> tst.add_dcm_spl('vls', rtn=True)    # same as .add_dcm_spl('vls', 'vls', rtn=True)/({'vls':'vls'}, rtn=True)
          cls vls
        0   1   1
        1   1   2
        2   1   3
        3   2   a
        4   2   b
        5   2   c
        :param args: in format (str_clm_spl, str_clm_new), ({str_clm_spl:str_clm_new},(str_clm_spl)
        :param prm: a regex to split the target cell
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: None if rtn
        """
        if 0 in self.clm:
            raise KeyError('a necessary column name 0 has been occupied, column 0 needs to be renamed')
        else:
            dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
            str_mpt, str_xpt = list(dct_rgs.keys())[0], list(dct_rgs.values())[0]
            dff = self.dts[str_mpt].str.split(prm, expand=True).stack()  # 分离为series形式
            dff.index = dff.index.codes[0]  # 保证index与self.ndx相符, 旧版本使用 = dff.index.labels[0]
            self.dts = concat([self.dts, dff], axis=1)
            self.drp_clm(str_mpt)
            self.rnm_clm({0: str_xpt})
            self.drp_dcm_ctt(str_xpt, '',)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_dcm_agg(self, lst_clm_grp, lst_clm_agg=None, *, clm_cls='_cty', clm_vls='_vls', spr=False, rtn=False):
        """
        add documents from aggregating, group by lst_clm_grp，lst_clm_mpt中的同性质列转行，实现缩列扩行.
        >>> tst = dfz([{'date': '1', 'vls_A': 1.0, 'vls_B': 3.0},
        >>>            {'date': '2', 'vls_A': 2.0, 'vls_B': 2.0}])
        >>> tst.typ_to_dtf()
        >>> tst.add_dcm_agg('date', ['vls_A', 'vls_B'], rtn=True)   # same as .add_dcm_agg('date', rtn=True)
            _cty date _vls
        0  vls_A    1    1
        1  vls_A    2    2
        3  vls_B    1    3
        4  vls_B    2    2
        :param lst_clm_grp: 用于汇总分组的列，即 grouping by and aggregating
        :param lst_clm_agg: 同性质的将要换位同一列的多列名组成的列表, 若表中除lst_clm_grp外均需处理则可为空
        :param clm_cls: 缩列后用于存放标志多列来源的分类列，列中内容为lst_clm_mpt，default _cty
        :param clm_vls: 缩列后用于存放多列数据值的唯一列名，default _vls
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: None
        """
        lst_dtf = []
        lst_clm_grp = lsz(lst_clm_grp).typ_to_lst(rtn=True)
        lst_clm_agg = [i for i in self.clm if i not in lst_clm_grp] if lst_clm_agg is None else \
            lsz(lst_clm_agg).typ_to_lst(rtn=True)
        for i_clm in lst_clm_agg:
            if i_clm in self.clm:
                lst_clm_grp_tmp = lsz(lst_clm_grp).typ_to_lst(rtn=True).copy()
                lst_clm_grp_tmp.append(i_clm)
                dtf_tmp = DataFrame(self.dts.values, columns=self.clm, index=lsz(i_clm).cpy_tal(self.len, rtn=True))
                dtf_tmp.index.name = clm_cls  # 设定index name 用于在reset_index环节直接导出成为column name
                dtf_tmp = dtf_tmp[lst_clm_grp_tmp]
                dtf_tmp.rename(columns={i_clm: clm_vls}, inplace=True)
                lst_dtf.append(dtf_tmp)  # 拼接用于for循环结束后的concat环节
        self.dts = concat(lst_dtf).reset_index()  # concat prm axis=0, reset_index prm drop=False
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts


class cllMixin(dcmMixin, clmMixin):
    """
    methods Mixin for collections.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(cllMixin, self).__init__(dts, lcn, spr=spr)

    def mrg_dtf(self, dtf_mrg, *args, spr=False, rtn=False, prm='outer'):
        """merge dataframe horizontally or vertically.
        >>> import pandas as pd
        >>> tst = dfz()
        >>> tst.mrg_dtf(pd.DataFrame([{"A":1}]), pd.DataFrame([{"A":2}]), prm='vrl', rtn=True)
           A
        0  1
        1  2
        >>> tst = dfz([{'A':1,'B':'a'},{'A':2,'B':'b'}])
        >>> tst.typ_to_dtf()
        >>> tst.mrg_dtf(pd.DataFrame([{'A':2,'C':'x'}]), 'A', rtn=True)
           A  B    C
        0  1  a  NaN
        1  2  b    x
        :param dtf_mrg: if prm is 'vrl': merge[self.dts,dtf_mrg,*args]; else: dataframe merged on the right side
        :param args: the column names for fitting in self.dts and dtf_mrg, len(lst_str_clm_on) in [1,2]
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: pandas.merge(how=str_how), in ['inner', 'outer', 'left', 'right','dpl'];
            str_how in ['dpl']: drop documents in self.dts matched in dtf_mrg
        :return: if rtn is True, return self.dts
        """
        if prm.lower() in ['vertical', 'vrl']:              # merging vertically
            lst_dtf = []
            if self.len > 0:
                lst_dtf.extend([self.dts])
            lst_dtf.extend([dtf_mrg])
            if args:
                lst_dtf.extend(lsz(args).typ_to_lst(rtn=True))
            self.dts = concat(lst_dtf, ignore_index=True)  # default keys=None, axis=0, sort=False
        elif prm.lower() in ['outer_index', 'outerindex', 'ndx']:         # merging by indexes
            self.dts = merge(self.dts, dtf_mrg, how='outer', left_index=True, right_index=True)
        else:
            lst_clm_on = lsz(args)
            lst_clm_on.typ_to_lst()
            lst_clm_on.cpy_tal(2, spr=True)         # 扩充用于匹配的字段名以满足left/right_on的需求
            if prm in ['inner', 'outer', 'left', 'right']:                      # traditional merging
                self.dts = merge(self.dts, dtf_mrg, how=prm,
                                 left_on=lsz(lst_clm_on[0]).typ_to_lst(rtn=True),
                                 right_on=lsz(lst_clm_on[1]).typ_to_lst(rtn=True))
            elif prm in ['dpl']:                    # merge and drop duplicates
                dtf_mrg_drp = dtf_mrg[lsz(lst_clm_on[1]).typ_to_lst(rtn=True)]  # 仅保留用于匹配的列，避免对self.dts的污染
                self.dts = merge(self.dts, dtf_mrg_drp, how='left',
                                 left_on=lsz(lst_clm_on[0]).typ_to_lst(rtn=True),
                                 right_on=lsz(lst_clm_on[1]).typ_to_lst(rtn=True),
                                 indicator=True)    # 使用自动定义的列'_merge'标记来源
                self.dts = self.dts.loc[self.dts['_merge'] != 'both', :].drop(['_merge'], axis=1)   # 删除标记来源行
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def t(self, *, spr=False, rtn=False):
        """
        对第一行第一列为坐标轴的dataframe进行转置.
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return:
        """
        str_clm_tgt = self.clm[0]
        dtmp = self.dts.T
        dtmp.columns = list(dtmp.iloc[0, :])
        dtmp = dtmp.iloc[1:, :]
        dtmp[str_clm_tgt] = dtmp.index
        self.dts = dtmp
        lst_clm_srt = list(self.clm)
        lst_clm_srt.remove(str_clm_tgt)
        lst_clm_srt = [str_clm_tgt] + lst_clm_srt
        self.srt_clm(lst_clm_srt, drp=False)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts


class tmsMixin(cllMixin):
    """
    methods Mixin for columns on type timeStamp.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(tmsMixin, self).__init__(dts, lcn, spr=spr)

    def typ_to_tms(self, *args, spr=False, rtn=False):
        """
        columns' type from others to datetime.datetime
        >>> tst = dfz([{'A':'1989-04-21'},{'A':'1989/4/21'}])
        >>> tst.typ_to_dtf()
        >>> tst.typ_to_tms('A','B')
                   A          B
        0 1989-04-21 1989-04-21
        1 1989-04-21 1989-04-21
        :param args: (columns) or (columns, newColumns) or ({columns: newColumns})
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: if rtn is True, return self.dts
        """
        lst_rgs = lsz(args).rgs_to_typ(prm='eql', rtn=True)
        lst_old, lst_new = lst_rgs[0], lst_rgs[1]
        for i in range(len(lst_old)):
            self.dts[lst_new[i]] = self.dts.apply(lambda x: dtz(x[lst_old[i]]).typ_to_dtt(rtn=True), axis=1)
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def tms_to_typ(self, *args, spr=False, rtn=False, prm='%Y-%m-%d', prn=True):
        """
        columns from timestamp to other types.
        >>> tst = dfz([{'A':'2019-04-21'},{'A':'2019/4/21'}])
        >>> tst.typ_to_dtf()
        >>> tst.tms_to_typ('A','yday',rtn=True)
             A
        0  111
        1  111
        :param args: ([columns],[types]) or ({column:type,...})
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: if type in 'str', define the format of the column
        :param prn: print info or not, default True
        :return: if rtn is True, return self.dts
        """
        if type(args[0]) is dict and len(args) == 1:
            lst_clm = lsz(list(args[0].keys())).typ_to_lst(rtn=True)
            lst_typ = lsz(list(args[0].values())).typ_to_lst(rtn=True)
        else:
            lst_clm = lsz(args[0]).typ_to_lst(rtn=True)
            lst_typ = lsz(args[1])
            lst_typ.typ_to_lst()
            lst_typ.cpy_tal(len(lst_clm), spr=True)
        for i in range(len(lst_clm)):
            if lst_typ[i] in ['w', 'm']:
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: dtz(x[lst_clm[i]]).dtt_to_prd(lst_typ[i], rtn=True, prn=prn), axis=1)
            else:
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: dtz(x[lst_clm[i]]).dtt_to_typ(lst_typ[i], prm, rtn=True), axis=1)
        self.dts_nit()
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_dcm_by_day(self, str_grb, str_tms, lst_prd=None):
        """
        对某列进行分类后对类中所有可能的取值进行时间列的行补足.
        >>> tst = dfz([{'cls':1,'date':'2019-01-01'},{'cls':1,'date':'2019-01-04'},])
        >>> tst.typ_to_dtf()
        >>> tst.add_dcm_by_day('cls','date')
        >>> tst.tms_to_typ('date','str')
        >>> tst.srt_dcm('date')
        >>> print(tst.dts)
           cls        date
        0    1  2019-01-01
        1    1  2019-01-02
        2    1  2019-01-03
        3    1  2019-01-04
        :param str_grb: 用于分类的列
        :param str_tms: 需要填充的时间列
        :param lst_prd: 补足的时间窗口
        :return: None
        """
        prcdtf = cllMixin(self.dts)
        lst_prd = [None, None] if lst_prd is None else lst_prd
        lst_prd[0] = prcdtf.dts[str_tms].min() if lst_prd[0] is None else lst_prd[0]
        lst_prd[1] = prcdtf.dts[str_tms].max() if lst_prd[1] is None else lst_prd[1]
        self.dts = DataFrame(dtz().lst_of_day(lst_prd, rtn=True), columns=[str_tms])
        for i_cll in list(prcdtf.dts[str_grb].unique()):
            dtf_cll = self.dts.copy()
            dtf_cll[str_grb] = i_cll
            dtf_cll = cllMixin(dtf_cll)
            dtf_cll.drp_dcm_ctt(str_tms, list(prcdtf.dts.loc[prcdtf.dts[str_grb] == i_cll, :][str_tms].unique()))
            prcdtf.mrg_dtf(dtf_cll.dts, prm='vrl')
        self.dts = prcdtf.dts


class pltMixin(cllMixin):
    """
    methods Mixin for plots and tables.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(pltMixin, self).__init__(dts, lcn, spr=spr)

    def _rp_cll_grb(self, lst_clm, lst_srt=None, prm_grb=-2, flt_drp=1):
        """
        drop cells on certain index of any columns for grouping by. 对变量列排序后，删除前flt_drp行特定列的值.
        >>> tst =dfz([
        >>>     {'A':'1','a':'x','B':100},
        >>>     {'A':'2','a':'x','B':110},
        >>>     {'A':'1','a':'y','B':150},
        >>>     {'A':'2','a':'y','B':100},
        >>> ])
        >>> tst.typ_to_dtf()
        >>> tst._rp_cll_grb('B', ['a','A'], -2, 1)
           A  a      B
        0  1  x    NaN
        1  2  x  110.0
        2  1  y    NaN
        3  2  y  100.0
        :param lst_clm: columns to be dropped
        :param lst_srt: columns to be sorted
        :param prm_grb: the target min column in lst_srt for grouping by. 分组后最小一级用于分组的变量所在lst_srt中的index
        :param flt_drp: 选择抛弃的行数, 顺时针计算
        :return: None
        """
        # 当存在多列排序时, 对不同汇总方式下的前n列进行赋空
        if lst_srt:
            self.srt_dcm(lst_srt)
        lst_srt = lsz(lst_srt).typ_to_lst(rtn=True)
        # 前几行的手工处理
        for i in range(flt_drp):
            for j in lst_clm:
                self.dts.loc[i, j] = np_nan
        # 当存在columns for grouping by时, 对每一组变量除datetime-like column外进行前n列的初始化
        if len(lst_srt) > 1:
            for i in range(self.len - 1):
                if self.dts.loc[i, lst_srt[prm_grb]] != self.dts.loc[i + 1, lst_srt[prm_grb]]:
                    for j in lst_clm:
                        for k in range(flt_drp):
                            self.dts.loc[i+k+1, j] = np_nan
        self.dts_nit()

    def add_clm_mom(self, *args, spr=False, rtn=False, prm=None, prm_grb=-2):
        """
        求环比变化率, add columns on month over month.
        >>> tst =dfz([{'A':'1','B':100},{'A':'2','B':110},{'A':'4','B':150},{'A':'3','B':100},])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_mom('B','b',prm='A')
           A    B       b
        0  1  100     NaN
        1  2  110  0.1000
        2  3  100 -0.0909
        3  4  150  0.5000
        >>> tst =dfz([{'A':'1','a':'x','B':100},{'A':'2','a':'x','B':110},{'A':'4','a':'x','B':150}])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_mom('B','b',prm=['a','A'],prm_grb=-2)
           A  a    B    b
        0  1  x  100  NaN
        1  2  x  110  0.1
        2  4  y  150  NaN   # 按照['a','A']的分组分别计算, 视prm_grb=-2为最细分组变量, prm[-1]通常为时间列
        :param args: target columns' name in format (x,) ,(oldColumn, newColumn), ({oldColumn: newColumn})
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: sorted documents by which columns before generating mom
        :param prm_grb: if multi-columns in prm, choose a colume index to restart mom
        :return: None
        """
        if prm:
            self.srt_dcm(prm)
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_clm = list(dct_rgs.keys())
        lst_new = [i+'_pop' for i in dct_rgs.keys()] if len(args) == 1 and type(args[0]) not in [dict] else \
            list(dct_rgs.values())
        for k in range(len(lst_clm)):
            for i in range(1, self.len):
                if self.dts.loc[i - 1, lst_clm[k]] > 0:
                    self.dts.loc[i, lst_new[k]] = round((self.dts.loc[i, lst_clm[k]] - self.dts.loc[i - 1, lst_clm[k]])
                                                        / self.dts.loc[i - 1, lst_clm[k]], 4)
        self.dts_nit()
        self._rp_cll_grb(lst_new, prm, prm_grb, 1)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_per(self, *args, spr=False, rtn=False, prm=None, drp=True):
        """
        add columns on each documents percents of each group. 根据某个分组汇总后计算每一行在该分组总量中的占比并形成新列.
        >>> tst =dfz([{'A':'a','B':100},{'A':'a','B':110},{'A':'b','B':150},{'A':'c','B':100},])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_per('A', 'B', 'b', prm=False)
           A    B       b
        0  a  100  0.4762
        1  a  110  0.5238
        2  b  150  1.0000
        3  c  100  1.0000
        :param prm: 用于分组的变量名列表
        :param args: 需要计算值的比例的变量名列表
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param drp: if drop template columns '_sum_x' or not, default True=drop.
        :return: None
        """
        prm = list(self.clm)[0] if prm is None else prm
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        dct_grb = dct_rgs.copy()
        for i in dct_grb.keys():
            dct_grb[i] = np_sum
        dtmp = self.dts.copy()
        dsrt = self.stt_clm(prm, dct_grb, prm=['_sum_'+i for i in dct_grb.keys()], rtn=True)
        self.dts = dtmp
        self.mrg_dtf(dsrt, prm, prm='left')
        for i in dct_rgs.keys():
            self.dts[dct_rgs[i]] = self.dts.apply(lambda x: round(x[i]/x['_sum_'+i], 4), axis=1)
        self.dts_nit()
        if drp:
            self.drp_clm(['_sum_'+i for i in dct_grb.keys()])
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_rll(self, *args, spr=False, rtn=False, prm=None, num=3, fnc='mean'):
        """
        add columns on rolling.
        >>> tst =dfz([
        >>>     {'A':'1','a':'x','B':100},
        >>>     {'A':'2','a':'x','B':110},
        >>>     {'A':'1','a':'y','B':150},
        >>>     {'A':'2','a':'y','B':100},
        >>> ])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_rll('B', 'b', num=2, fnc='mean', prm=['a','A'], rtn=True)
           A  a    B      b
        0  1  x  100    NaN
        1  2  x  110  105.0
        2  1  y  150    NaN
        3  2  y  100  125.0
        :param args: in format ([columns for rolling],[new columns for the rolling columns])
        :param num: window's width
        :param fnc: rolling method, in ['sum', 'mean', 'std', 'max', 'min']
        :param prm: columns for sorting before rolling, the last cell would be a datetime-like column
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :return: None
        """
        if prm:
            self.srt_dcm(prm)
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_clm = list(dct_rgs.keys())
        lst_new = [i+'_rll' for i in dct_rgs.keys()] if \
            len(args) == 1 and type(args[0]) not in [dict] else list(dct_rgs.values())
        # 摘取需要rolling的部分单独进行计算并在完成计算后使用index横向拼接到原数据集
        rll = cllMixin(self.dts[lst_clm].copy())
        if fnc in [sum, 'sum']:
            rll.dts = rll.dts.loc[:, lst_clm].rolling(num).sum()
        elif fnc in ['mean']:
            rll.dts = rll.dts.loc[:, lst_clm].rolling(num).mean()
        elif fnc in ['std']:
            rll.dts = rll.dts.loc[:, lst_clm].rolling(num).std()
        elif fnc in ['max']:
            rll.dts = rll.dts.loc[:, lst_clm].rolling(num).max()
        elif fnc in ['min']:
            rll.dts = rll.dts.loc[:, lst_clm].rolling(num).min()
        rll.rnm_clm(lst_clm, lst_new)
        self.mrg_dtf(rll.dts, prm='outerIndex')
        self._rp_cll_grb(lst_new, prm, -2, num-1)
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts

    def add_clm_nrm(self, *args, spr=False, rtn=False, prm=None, ltr=False):
        """
        add columns for normalization.
        >>> tst = dfz([{'A': 1, 'B': 'a', 'vls': 100}, {'A': 1, 'B': 'a', 'vls': 110}, {'A': 1, 'B': 'c', 'vls': 120}])
        >>> tst.typ_to_dtf()
        >>> tst.add_clm_nrm('vls', prm=['A'])
           A  B  vls  vls_nrm
        0  1  a  100     -1.0
        1  1  a  110      0.0
        2  1  c  120      1.0
        :param args: ([columns for normalization],[new columns name])
        :param spr: let self = self.dts
        :param rtn: default False, return None
        :param prm: list of columns to group by
        :param ltr: 如果某个值过大, 则修正为三倍标准差, 默认False
        :return: None
        """
        dct_rgs = lsz(args).rgs_to_typ(prm='dct', rtn=True)
        lst_clm = list(dct_rgs.keys())
        lst_new = [i+'_nrm' for i in dct_rgs.keys()] if \
            len(args) == 1 and type(args[0]) not in [dict] else list(dct_rgs.values())
        if prm is None:
            self.dts['_grb'] = 1
            self.dts_nit()
            self.srt_clm('_grb', drp=False)
        lst_grb = lsz(prm).typ_to_lst(rtn=True) if prm else [list(self.clm)[0]]     # 用于分组的列
        # 计算分组的均值
        avg = dfz(self.dts.copy())
        avg.stt_clm(lst_grb, lst_clm, [np_mean])
        avg.rnm_clm({i: '_'+i+'_avg' for i in lst_clm})
        # 计算分组的标准差
        std = dfz(self.dts.copy())
        std.stt_clm(lst_grb, lst_clm, [np_std])
        std.rnm_clm({i: '_'+i+'_std' for i in lst_clm})
        avg.mrg_dtf(std.dts, lst_grb, prm='outer')
        # 横向拼接和学生标准化
        self.mrg_dtf(avg.dts, lst_grb, prm='outer')
        if ltr:
            for i in range(len(lst_clm)):
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: x['_'+lst_clm[i]+'_avg'] + 3*x['_'+lst_clm[i]+'_std'] if
                    x[lst_clm[i]] - x['_'+lst_clm[i]+'_avg'] > 3*x['_'+lst_clm[i]+'_std'] else
                    x[lst_clm[i]], axis=1
                )
                self.dts[lst_clm[i]] = self.dts.apply(
                    lambda x: x['_'+lst_clm[i]+'_avg'] - 3*x['_'+lst_clm[i]+'_std'] if
                    x[lst_clm[i]] - x['_'+lst_clm[i]+'_avg'] < -3*x['_'+lst_clm[i]+'_std'] else
                    x[lst_clm[i]], axis=1
                )
        for i in range(len(lst_clm)):
            self.dts[lst_new[i]] = self.dts.apply(
                lambda x: round((x[lst_clm[i]] - x['_'+lst_clm[i]+'_avg']) / x['_'+lst_clm[i]+'_std'], 6), axis=1
            )
        self.dts_nit()
        self.drp_clm(['_grb'])
        self.drp_clm(['_'+i+'_std' for i in lst_clm])
        self.drp_clm(['_'+i+'_avg' for i in lst_clm])
        if spr:
            self.spr_nit()
        if rtn:
            return self.dts


class dfz(tmsMixin, pltMixin):
    """
    dataFrame operating class with slots.
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(dfz, self).__init__(dts, lcn, spr=spr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
spider operation.
@author: zoharslong
"""
from time import sleep
from random import randint
from re import findall as re_find, sub as re_sub
from numpy import nan as np_nan, count_nonzero as np_count
from pyzohar.sub_slt_bsc.dfz import dtz, dfz


class spz_nit(dfz):
    """
    web spider class initiation
    >>> dct_ppc = {
    >>>     'ndx_rgx':{'_id':'(\d+)'},  # 指定详情页url替换规则: {用于搜索详情页的ID: url_ctt中需要替换的正则表达式}
    >>>     'ndx':['_id','...'],        # 最终入库时的唯一组合键, 必然包含上一行的唯一识别ID
    >>>     'clm_typ':{}                # 其他常规列处理参数
    >>> }
    >>> spd = spz_nit(lcn={
    >>>     'url_lst': 'https://sz.ke.com/chengjiao/start/pg(\d+)/',    # crt_url_lst使(\d+)递增生成url, start为占位符
    >>>     'url_ctt': 'https://sz.ke.com/chengjiao/c(\d+)/pg1/',       # crt_url_ctt使ppc.ndx_rgx.key替换(\d+)生成url
    >>>     'prx':'auto', 'prx_tkn': {'neek':'', 'appkey':''},          # 开启极光代理
    >>>     'hdr':{},                                                   # 虚构请求头
    >>>     'ppc':dct_ppc})
    >>> spd.spd_bch(None, prm='lst', pg_max=9)  # 爬取列表页, 最高至第八页, 数据处理函数选择None
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        """
        more param of self is available, in [clm_x, clm_y, _x, _y, mdl]
        :param dts: self.dts, premier than dfz_mpt
        :param lcn: self.lcn, premier than dfz_mpt
        :param spr: let self = self.dts
        """
        super(spz_nit, self).__init__(dts, lcn, spr=spr)

    def crt_url_lst(self, pg_max=100):
        """
        create url on list pages. 对列表类的url爬虫, 做url页面的序号递增.
        :param pg_max: 最大页面控制数
        :return: None
        """
        if 'url' not in self.lcn.keys() or re_find(self.lcn['url_ctt'], self.lcn['url']):   # 初次定义url
            self.lcn['url'] = self.lcn['url_lst'].replace('(\d+)', '1')
        else:
            now = int(re_find(self.lcn['url_lst'].replace('?', '\?'), self.lcn['url'])[0])
            pgs = now + 1
            if pgs >= pg_max:   # 最大爬取页面控制线
                self.lcn.pop('url')
            else:               # 正常情况下的操作
                self.lcn['url'] = self.lcn['url_lst'].replace('(\d+)', str(pgs))   # 在已有url的基础上对url做递增处理

    def crt_url_ctt(self, prd_gap=15, srt=None):
        """
        create url on content pages. 对内容类的url爬虫, 从mongodb中搜索对应的详情页id拼接形成新的内容页url.
        @prd_gap: 设定的时间间隔, 当目标estateID最后一次运行爬虫距今时差大于这个值时, 执行爬虫
        @return: None
        """
        srt = {'cnt': -1} if srt is None else srt
        edg = dtz('now')
        edg.shf(-prd_gap)
        edg = edg.dtt_to_typ('str', '%Y-%m-%d %H:%M:%S', rtn=True)
        eid = False
        try:
            eid = self._myCln.aggregate([
                {'$match': {'$or': [
                    {'__time_ctt': {'$exists': False}},
                    {'__time_ctt': {'$lte': edg}},
                    {'__time_ctt': {'$in': [None, '', np_nan]}}
                ]}},
                {'$group': {'_id': '$' + list(self.lcn['ppc']['ndx_rgx'].keys())[0], 'cnt': {'$sum': 1}}},
                {'$sort': srt}]).next()['_id']
        except StopIteration:   # 当无法查找到目标时pymongo会报Error: StopIteration, 此时将不再产生有效的url
            pass
        finally:
            if eid:
                if 'url' not in self.lcn.keys() or re_find(self.lcn['url_lst'], self.lcn['url']):
                    self.lcn['url'] = self.lcn['url_ctt'].replace(list(self.lcn['ppc']['ndx_rgx'].values())[0], eid)
                else:
                    rpc = re_find(self.lcn['url_ctt'], self.lcn['url'])[0]
                    self.lcn['url'] = self.lcn['url'].replace(rpc, eid)
            elif 'url' in self.lcn.keys():   # 如果库中已不存在有效的eid
                self.lcn.pop('url')

    def spd_xpt(self, fnc_sop=None, *, rty=2, frc=False, rtn=True):
        """
        spider export from html files.
        @param fnc_sop: a special function to deal with target html
        @param rty: repeat running times
        @param frc: force to change proxy ip
        @param rtn: return spider is alright or not
        @return: in [True, False] for [got target, cannot got anything]
        """
        prc, bch = True, 0
        while prc and bch <= rty:           # 对self.api_run获得的信息进行有效性判断后发现无效时的循环重复
            if bch == rty and rty >= 2:     # 当最后一次循环的时候强制更换proxy
                self.api_run(frc=True, prm='get')
            else:                           # 其他时候不强制刷新一次proxy
                self.api_run(frc=frc, prm='get')
            if fnc_sop is None:             # 当未指定soup - xml处理函数时在此处中断运行
                raise KeyError('stop: need a fnc_sop.')
            self.dts = fnc_sop(self)        # 自定义的网页数据摘取过程
            self.drp_dcm_dpl(self.lcn['ppc']['ndx'])
            if self.len > 0:                # 此时代表经过fnc_sop处理后, 有效的数据已经产生了, 循环结束
                self.mng_xpt(self.lcn['ppc']['ndx'])
                prc = False                 # 网页定制化处理有数据且成功入库后, 终止循环
                if rtn:
                    return True
            else:
                self.api_prx(frc=True)
                if rtn and bch == rty:      # 当第三次循环仍旧未能得到理想结果时, return False
                    return False
            bch += 1
            sleep(1 + randint(0, 2))

    def spd_bch(self, fnc_sop=None, *, prm='lst', srt=None, pg_max=100):
        """
        spider batch. 循环运作直到self.spd_xpt返回False停止.
        :param fnc_sop: a special function to deal with target html
        :param prm: in ['list', 'lst', 'content', 'ctt']
        :param srt: only when prm in ['ctt'], define the sort of id, default {'cnt': -1}
        :param pg_max: default 100
        :return: None
        """
        self.lcn['url_lst'] = self.lcn['url_lst'].replace('start/', '')     # 使用本方法即可排除需要二次循环的情况
        prc = True
        while prc:                              # 决定是否终止整个爬虫循环, 每个循环代表一个新的url被执行
            if prm in ['list', 'lst']:
                self.crt_url_lst(pg_max)        # 创造本次爬取的url
                if 'url' not in self.lcn.keys():    # 当self.crt_url未能产生新的url时终止batch running
                    print('stop: url not exist.')
                    break
                else:
                    prc = self.spd_xpt(fnc_sop)     # 爬取页并分离其中的数据导入mongodb, 当爬取列表页连续三次失败时终止batch running
            elif prm in ['content', 'ctt']:
                self.crt_url_ctt(srt=srt)
                if 'url' not in self.lcn.keys():    # 当self.crt_url未能产生新的url时终止batch running
                    print('stop: url not exist.')
                    break
                else:
                    self.spd_xpt(fnc_sop, rtn=False)
            sleep(2 + randint(0, 5))

    def spd_bch_whl_swh(self, fnc_sop=None, *, srt=None, pg_max=100, lst_bch=None):
        """
        专用定制函数 - 由于贝壳每个筛选下只能显示100页，因此对各个行政区分别运行列表页爬虫; 本质就是self.spd_bch
        :param fnc_sop: default sop_est_shd_bke_lst
        :param srt: default {'cnt': -1}
        :param pg_max: max pages, default 100
        :param lst_bch: a list of different districts, default for shenzhen beike
        :return: None
        """
        lst_bch = [
            'longgangqu', 'longhuaqu', 'luohuqu', 'futianqu', 'dapengxinqu',
            'yantianqu', 'nanshanqu', 'baoanqu', 'guangmingqu', 'pingshanqu'
        ] if lst_bch is None else lst_bch   # default shenzhen beike
        if lst_bch[0] != 'start':
            lst_bch = ['start'] + lst_bch
        if lst_bch[-1] != 'end':
            lst_bch += ['end']
        for i in range(len(lst_bch) - 1):
            try:
                self.lcn.pop('url')  # self.lcn中不存在url以保证初始化状态
            except KeyError:
                pass
            if lst_bch[i] in ['start']:
                print('info: page series switch from %s to %s.' % (lst_bch[i], lst_bch[i+1]))
                self.lcn['url_lst'] = re_sub(lst_bch[i], lst_bch[i + 1], self.lcn['url_lst'])
            elif re_find(lst_bch[i], self.lcn['url_lst']):
                self.spd_bch(fnc_sop, prm='lst', srt=srt, pg_max=pg_max)
                print('info: page series switch from %s to %s.' % (lst_bch[i], lst_bch[i+1]))
                self.lcn['url_lst'] = re_sub(lst_bch[i], lst_bch[i + 1], self.lcn['url_lst'])

    def ltr_spd_for_dcm(self, flt_dcm=2, mtd='lte', prm='drop', nit=True):
        """
        因为现有文档数量的原因, 通过修改时间的形式对是否运作某一个文档的爬虫进行预设, 将成交数量较少的楼盘排除出爬虫序列
        默认情况为将文档数量少于等于2的ID排除出爬虫序列不再爬取
        :param flt_dcm: 文档数量判断
        :param mtd: 大于还是小于上述文档数量边界, 默认'lte' in [‘lte’,'gt']
        :param prm: 对满足上述条件的文档执行什么操作, 默认放弃爬取'drop' in ['drop','keep']
        :param nit: 是否对'__time_ctt'为空的行也进行修改操作, 默认True
        :return: None
        """
        str_dtz = dtz('now').dtt_to_typ(str_fmt='%Y-%m-%d %H:%M:%S', rtn=True) if prm in ['drop', 'drp'] else \
            dtz(dtz('now').shf(-30, rtn=True)).dtt_to_typ(str_fmt='%Y-%m-%d %H:%M:%S', rtn=True)
        str_key = list(self.lcn['ppc']['ndx_rgx'].keys())[0]
        self.mng_mpt({}, [str_key, '__time', '__time_ctt'])
        self.typ_to_dtf()
        if not nit:
            self.drp_dcm_na('__time_ctt')
        self.stt_clm(str_key, '__time', [np_count])
        if mtd in ['lte']:
            self.dts = self.dts.loc[self.dts['__time'] <= flt_dcm]
        elif mtd in ['lt']:
            self.dts = self.dts.loc[self.dts['__time'] < flt_dcm]
        elif mtd in ['gte']:
            self.dts = self.dts.loc[self.dts['__time'] >= flt_dcm]
        elif mtd in ['gt']:
            self.dts = self.dts.loc[self.dts['__time'] > flt_dcm]
        for i in range(self.len):  # 更新爬虫时间
            if nit:
                self.ltr_cln_dcm({str_key: self.dts.loc[i, str_key]}, {'$set': {'__time_ctt': str_dtz}})
            else:
                self.ltr_cln_dcm({str_key: self.dts.loc[i, str_key], '__time_ctt': {'$exists': True}},
                                 {'$set': {'__time_ctt': str_dtz}})


class spz(spz_nit):
    """
    web spider
    """
    def __init__(self, dts=None, lcn=None, *, spr=False):
        super(spz, self).__init__(dts, lcn, spr=spr)
